#include "../std.h"
#include <globus_common.h>
#include <globus_ftp_client.h>
#include <globus_ftp_client_plugin.h>
/* #include <globus_ftp_client_debug_plugin.h> */
#include <globus_gass_copy.h>
#include <iostream>

#include "../misc/checkfile.h"
#include "../misc/stringtoint.h"
#include "../misc/log_time.h"

#include "check.h"

#define GLOBUS_ERROR GLOBUS_ERROR_NO_INFO
globus_object_t globus_error_not_found = globus_object_static_initializer( \
            GLOBUS_OBJECT_TYPE_PRINTABLE,GLOBUS_ERROR_BASE_PROTOTYPE);
/*          GLOBUS_ERROR_TYPE_BASE,GLOBUS_ERROR_BASE_PROTOTYPE); */

/*
pthread_mutex_t global_check_lock = PTHREAD_MUTEX_INITIALIZER;
*/

typedef struct {
  globus_gass_copy_callback_t cb;
  void * arg;
  globus_gass_copy_handle_t * handle;
  globus_object_t * error;
} callback_args_t;

void* callback_call(void* args) {
  (*(((callback_args_t*)args)->cb)) (
    ((callback_args_t*)args)->arg,
    ((callback_args_t*)args)->handle,
    ((callback_args_t*)args)->error
  );
/*    (*cb)(arg,a_handle,GLOBUS_SUCCESS);  */
  free(args);
  globus_thread_exit(NULL); return NULL;
}

globus_result_t callback_call_thread(globus_gass_copy_handle_t *handle,globus_gass_copy_callback_t cb,void *arg) {
  globus_thread_t callback_thread;
  callback_args_t *args;
  args=(callback_args_t*)malloc(sizeof(callback_args_t));
  if(args == NULL) {
    olog<<"Memory allocation error"<<endl; return GLOBUS_ERROR;
  };
  args->cb=cb;
  args->arg=arg;
  args->handle=handle;
  args->error=GLOBUS_SUCCESS;
  if(globus_thread_create(&callback_thread,NULL,&callback_call,args) != 0) {
    olog<<"Failed to start new thread"<<endl;
    free(args);
    return GLOBUS_ERROR;
  };
  return GLOBUS_SUCCESS;
}

char* remove_spaces(const char* url) {
  char* url_ = (char*)malloc(strlen(url)*3+1);
  if(url_ == NULL) return NULL;
  const char* s = url;
  char* s_space;
  int l=0;
  for(;;) {
    s_space=strchr(s,' ');
    if(s_space == NULL) {
      int ll=strlen(s);
      memcpy(url_+l,s,ll);
      url_[l+ll]=0;
      break;
    };
    memcpy(url_+l,s,s_space-s);
    l+=(s_space-s);
    url_[l]='%'; l++; url_[l]='2'; l++; url_[l]='0'; l++; 
    s=s_space+1; 
  };
  return url_;
}

static globus_result_t check_file_url(const char* url,uid_t uid) {
  const char* s = url +6;
  if((*s) == '/') { 
    s++;
    s = strchr(s,'/');
    if((*s) != '/') {  /* bad url */
      olog<<"URL "<<url<<" is bad"<<endl;
      return GLOBUS_ERROR;
    };
  }
  else { s--; };
  int res;
  res=check_file_access(s,uid,0,(gid_t)(-1));
  if(res == -1) { /* no such file */
    olog<<"File "<<s<<" does not exist or is not a file (most probably)"<<endl;
    return GLOBUS_ERROR_LOCAL_FILE_NOT_FOUND;
  };
  if(res != 0) {
    olog<<"File "<<s<<" is not accessible"<<endl;
    return GLOBUS_ERROR;
  };
  return GLOBUS_SUCCESS;
}

globus_result_t globus_gass_copy_register_url_to_url_check(
    globus_gass_copy_handle_t *a_handle,
    const char* s_url,globus_gass_copy_attr_t *s_attr,
    const char* d_url,globus_gass_copy_attr_t *d_attr,
    globus_gass_copy_callback_t cb,void* arg) {
  globus_result_t res;

  if(!strncmp(s_url,"file:/",6)) {
    if(check_file_url(s_url,get_user_id()) != GLOBUS_SUCCESS)
                                                   return GLOBUS_ERROR;
    /* check if that is file to file download */
    if(!strncmp(d_url,"file://",7)) {
      /* let gass_copy handle copy */   


    };
  };
  if(!strncmp(s_url,"link://",7)) {
    /* only link to file allowed */
    if(strncmp(d_url,"file://",7)) {
      olog<<"Cannot make link "<<d_url<<" to remote file "<<s_url<<endl;
      return GLOBUS_ERROR;
    };
    const char* s_file = s_url+7;
    const char* d_link = d_url+7;
    if(symlink(s_file,d_link) == -1) {
      olog<<"Failed making link "<<d_link<<" to file "<<s_file<<endl;
      return GLOBUS_ERROR;      
    };
    /* call callback with success from separate thread to avoid deadlock */
    return callback_call_thread(a_handle,cb,arg);
  };
  /* remove spaces */
  char* s_url_ = remove_spaces(s_url);
  if(s_url_ == NULL) return GLOBUS_ERROR;
  char* d_url_ = remove_spaces(d_url);
  if(d_url_ == NULL) { free(s_url_); return GLOBUS_ERROR; };
  res = globus_gass_copy_register_url_to_url(
                    a_handle,(char*)s_url_,s_attr,(char*)d_url_,d_attr,cb,arg);
  free(d_url_);
  free(s_url_);
  return res;
}

struct ftp_args_s {
  globus_mutex_t ftp_mutex;
  globus_cond_t ftp_cond;
  int ftp_flag;
  globus_byte_t buf[4096];
};

void ftp_finished(void* arg,globus_ftp_client_handle_t *handle,globus_object_t* error) {
   struct ftp_args_s *f_args = (struct ftp_args_s *)arg;
   globus_mutex_lock(&(f_args->ftp_mutex));
   if(error == GLOBUS_SUCCESS) { f_args->ftp_flag=1; }
   else { f_args->ftp_flag=2; };
   globus_cond_signal(&(f_args->ftp_cond));
   globus_mutex_unlock(&(f_args->ftp_mutex));
};

void ftp_read(void* arg,globus_ftp_client_handle_t *handle,globus_object_t* error,globus_byte_t* buffer,globus_size_t length,globus_off_t offset,globus_bool_t eof) {
  struct ftp_args_s *f_args = (struct ftp_args_s *)arg;
  if(!eof) globus_ftp_client_register_read(handle,f_args->buf,16,ftp_read,f_args);
}

/* do not download anything - only check if file is downloadable */
globus_result_t globus_gass_copy_url_access_check(
    const char* s_url,globus_gass_copy_attr_t *s_attr) {

  if((strncmp(s_url,"file:/",6)==0) || (strncmp(s_url,"link:/",6)==0)) {
    const char* s = s_url +6;
    if((*s) == '/') { 
      s++;
      s = strchr(s,'/');
      if((*s) != '/') {  /* bad url */
        olog<<"URL "<<s_url<<" is bad"<<endl;
        return GLOBUS_ERROR;
      };
    };
    if(check_file_url(s_url,get_user_id()) != GLOBUS_SUCCESS)
                                                     return GLOBUS_ERROR;
    return GLOBUS_SUCCESS;
  }
  else if(!strncmp(s_url,"https://",8)) {
    /* do not know how to check access */
    olog<<"acsess check: check not supported: "<<s_url<<endl;
    return GLOBUS_SUCCESS;
  }
  else if( (!strncmp(s_url,"gsiftp://",9))  ||
           (!strncmp(s_url,"ftp://",6)) ) { 
// connect to the site and download first byte of file
    struct ftp_args_s f_args;
/*
    pthread_mutex_lock(&global_check_lock);
*/

    globus_result_t ftp_res;
    globus_ftp_client_plugin_t plg;
    globus_ftp_client_handle_t ftp_h;
    globus_ftp_client_handleattr_t ftp_attr;
    globus_ftp_client_operationattr_t ftp_opattr;

    globus_ftp_client_handleattr_init(&ftp_attr);
    globus_ftp_client_handleattr_set_cache_all(&ftp_attr,GLOBUS_TRUE);
/*
    globus_ftp_client_debug_plugin_init(&plg,stderr,s_url);
    globus_ftp_client_handleattr_add_plugin(&ftp_attr,&plg);
*/
    globus_ftp_client_handle_init(&ftp_h,&ftp_attr);
    globus_ftp_client_operationattr_init(&ftp_opattr);

    globus_mutex_init(&(f_args.ftp_mutex),GLOBUS_NULL);
    globus_cond_init(&(f_args.ftp_cond),GLOBUS_NULL);
    f_args.ftp_flag=0;
    ftp_res=globus_ftp_client_partial_get(&ftp_h,s_url,
                  &ftp_opattr,GLOBUS_NULL,0,1,&ftp_finished,&f_args);
//    ftp_res=globus_ftp_client_exists(&ftp_h,f_args.url,
//                  &ftp_opattr,&ftp_finished,f_args);
    if(ftp_res == GLOBUS_SUCCESS) {
      globus_ftp_client_register_read(&ftp_h,f_args.buf,16,&ftp_read,&f_args);
      globus_mutex_lock(&(f_args.ftp_mutex));
      while(f_args.ftp_flag == 0) {
        globus_cond_wait(&(f_args.ftp_cond),&(f_args.ftp_mutex));
      };
      if(f_args.ftp_flag != 1) ftp_res=GLOBUS_ERROR;
      globus_mutex_unlock(&(f_args.ftp_mutex));
    }
    else {
      char* tmp=globus_object_printable_to_string(globus_error_get(ftp_res));
      olog<<tmp<<endl; free(tmp);
    };
/*
    globus_ftp_client_handleattr_remove_plugin(&ftp_attr,&plg);
*/
    globus_ftp_client_handle_destroy(&ftp_h);
    globus_ftp_client_handleattr_destroy(&ftp_attr);
    globus_ftp_client_operationattr_destroy(&ftp_opattr);
/*
    globus_ftp_client_debug_plugin_destroy(&plg);
*/

    globus_cond_destroy(&(f_args.ftp_cond));
    globus_mutex_destroy(&(f_args.ftp_mutex));

/*
  pthread_mutex_unlock(&global_check_lock);
*/

    return ftp_res;
  }
  else { /* unchekable protocols */
    olog<<"acsess check: uncheckable proto: "<<s_url<<endl;
    /* call callback with success from separate thread to avoid deadlock */
    return GLOBUS_SUCCESS;
  };  
}

